# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Validate build log for additional vulnerabilities."""

import os
import argparse
import re
import sys
from pathlib import Path
from azureml.assets.util import logger


def validate_py_version(build_log_file_name, build_log_content):
    """Validate Python version.

    Args:
        build_log_file_name (str): Build log file name.
        build_log_content (str): Build log content

    Returns:
        int: Number of errors.
    """
    py38_match = re.search("python=3.8", build_log_content)

    if py38_match:
        logger.log_error(f"{build_log_file_name}: python=3.8 found in build log. "
                         f"Python 3.8 is now deprecated. Please use a newer Python version.")
        return 1

    return 0


def validate_ubuntu_version(build_log_file_name, build_log_content):
    """Validate Ubuntu version.

    Args:
        build_log_file_name (str): Build log file name.
        build_log_content (str): Build log content

    Returns:
        int: Number of errors.
    """
    ubuntu2004_match = re.search(r"FROM\s+(.*20\.04.*|.*2004.*)", build_log_content)

    if ubuntu2004_match:
        logger.log_error(f"{build_log_file_name}: Ubuntu 20.04 reference found in build log. "
                         f"Ubuntu 20.04 is nearing or has reached EOL. Please create an environment "
                         f"based on a newer Ubuntu version.")
        return 1

    return 0


def validate_azureml_sdk_dependencies(build_log_file_name, build_log_content):
    """Validate AzureML SDK dependencies.

    Args:
        build_log_file_name (str): Build log file name.
        build_log_content (str): Build log content

    Returns:
        int: Number of errors.
    """
    azure_ai_ml_match = re.search("azure-ai-ml", build_log_content)
    azureml_core_match = re.search("azureml-core", build_log_content)

    if azure_ai_ml_match and azureml_core_match:
        logger.log_error(f"{build_log_file_name}: azure-ai-ml and azureml-core depedencies both found in build log. "
                         f"Please use only one version of the AzureML SDK.")
        return 1

    return 0


def validate_build_logs(build_logs_dir):
    """Validate environment build logs.

    Args:
        build_logs_dir (Path): Directory of environment build logs.

    Returns:
        bool: True if build logs were successfully validated, otherwise False.
    """
    error_count = 0

    for build_log_file_name in os.listdir(build_logs_dir):

        build_log_file_path = os.path.join(build_logs_dir, build_log_file_name)
        print(f"Validating {build_log_file_name} for additional vulnerabilities")

        with open(build_log_file_path, "r") as f:
            build_log_content = f.read()
            error_count += validate_py_version(build_log_file_name, build_log_content)
            error_count += validate_ubuntu_version(build_log_file_name, build_log_content)
            error_count += validate_azureml_sdk_dependencies(build_log_file_name, build_log_content)

    return error_count == 0


if __name__ == '__main__':
    # Handle command-line args
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--build-logs-dir", required=True, type=Path,
                        help="Directory of build logs")
    args = parser.parse_args()

    if not os.path.isdir(args.build_logs_dir):
        print(f"No build logs found since {args.build_logs_dir} does not exist")
        sys.exit(0)

    # Validate build logs
    success = validate_build_logs(build_logs_dir=args.build_logs_dir)

    if success:
        print("No additional vulnerabilities found in build logs")
    else:
        sys.exit(1)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""SKU utils."""

import json
import requests
from azureml.assets.util import logger
from azureml.assets.util.util import retry
from azure.identity import AzureCliCredential


all_sku_details = None
SKU_DETAILS_URI = (
    "https://management.azure.com/subscriptions/{}/"
    "providers/Microsoft.MachineLearningServices/locations/{}"
    "/vmSizes?api-version=2021-01-01&expandChildren=true"
)


@retry(3)
def get_all_sku_details(credential: AzureCliCredential, subscription_id: str, location: str = "eastus"):
    """Return all sku details.

    return response =>
    {
        "Standard_A1_v2": {
            "name": "Standard_A1_v2",
            "family": "standardAv2Family",
            "vCPUs": 1,
            "gpus": 0,
            "osVhdSizeMB": 1047552,
            "maxResourceVolumeMB": 10240,
            "memoryGB": 2.0,
            ....
        },
        ....
        ....
    }

    Args:
        credential (AzureCliCredential): Credential to generate token for the request
        subscription_id (str): Subscription ID to check details in
        location (str): location to query SKU details for. Default is set to eastus
    """
    global all_sku_details
    if all_sku_details is None:
        vmSizes = SKU_DETAILS_URI.format(subscription_id, location)
        token = credential.get_token("https://management.azure.com/.default")
        headers = {"Authorization": f"Bearer {token.token}"}
        response = requests.get(vmSizes, headers=headers)
        status_code = response.status_code
        content = response.content
        if status_code != 200:
            raise Exception(f"Unsuccessful requst. Response : {response}")
        sku_details_list = json.loads(content).get("amlCompute", [])
        all_sku_details = {sku_details["name"]: sku_details for sku_details in sku_details_list}

    return all_sku_details


def get_sku_details(credential: AzureCliCredential, SKU: str, subscription_id: str, location: str = "eastus"):
    """Get sku details.

    Args:
        credential (AzureCliCredential): Credential to generate token for the request
        SKU (str): SKU to fetch detail of
        subscription_id (str): Subscription ID to check details in
        location (str): location to query SKU details for. Default is set to eastus
    """
    global all_sku_details
    if all_sku_details is None:
        logger.print(f"Fetching all sku details for subscription: {subscription_id} and location: {location}")
        all_sku_details = get_all_sku_details(
            credential,
            subscription_id,
            location,
        )

    return all_sku_details.get(SKU, None)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Tools to maintain asset source files."""

import argparse
import shutil
from pathlib import Path

import azureml.assets as assets
import azureml.assets.util as util
from azureml.assets.util import logger

ASSET_COUNT = "asset_count"
DELETED_COUNT = "deleted_count"


def list_assets(args: argparse.Namespace):
    """List assets.

    Args:
        args (argparse.Namespace): Args from argparse.
    """
    # Find assets under input dir
    asset_list = []
    for asset_config in util.find_assets(args.input_dirs, args.asset_config_filename):
        asset_list.append(asset_config.partial_name)
    asset_list.sort()

    # Write to file or stdout
    if args.output_file:
        with open(args.output_file, "w") as f:
            f.write("\n".join(asset_list))
    else:
        print("\n".join(asset_list))

    # Set variables
    logger.set_output(ASSET_COUNT, len(asset_list))


def delete_assets(args: argparse.Namespace):
    """Delete assets that are not in the retention file.

    NOTE: This is generally meant to be run against the release branch, where there's
    no risk of deleting shared files like component source code.

    Args:
        args (argparse.Namespace): Args from argparse.
    """
    # Read retention file
    with open(args.retention_file) as f:
        retention_list = f.read().splitlines()
        print(f"Read {len(retention_list)} asset(s) from retention file")

    # Find assets under input dir
    asset_count = 0
    deleted_count = 0
    for asset_config in util.find_assets(args.input_dirs, args.asset_config_filename):
        asset_count += 1
        if asset_config.partial_name not in retention_list:
            # Get common directory, in case asset config file isn't at the root
            common_dir, _ = util.find_common_directory(asset_config.release_paths)

            # Delete asset
            if args.dry_run:
                logger.print(f"Would delete {asset_config.partial_name} from {common_dir}")
            else:
                try:
                    logger.print(f"Deleting {asset_config.partial_name} from {common_dir}")
                    shutil.rmtree(common_dir)
                    deleted_count += 1
                except Exception as e:
                    logger.log_warning(f"Failed to delete {common_dir}: {e}")

    # Set variables
    logger.set_output(ASSET_COUNT, asset_count)
    logger.set_output(DELETED_COUNT, deleted_count)


if __name__ == '__main__':
    # Quick function to convert comma-separated arg to Path list
    def _list_path(value: str):
        return [Path(d) for d in value.split(",")]

    # Handle command-line args
    parser = argparse.ArgumentParser()
    shared_parser = argparse.ArgumentParser(add_help=False)
    shared_parser.add_argument("-i", "--input-dirs", type=_list_path, required=True,
                               help="Comma-separated list of directories containing assets")
    shared_parser.add_argument("-a", "--asset-config-filename", default=assets.DEFAULT_ASSET_FILENAME,
                               help="Asset config file name to search for")
    subparsers = parser.add_subparsers()

    parser_list = subparsers.add_parser("list", help="List assets", parents=[shared_parser])
    parser_list.set_defaults(func=list_assets)
    parser_list.add_argument("-o", "--output-file", type=Path,
                             help="File to which asset names will be written")

    parser_delete = subparsers.add_parser("delete", help="Delete assets", parents=[shared_parser])
    parser_delete.set_defaults(func=delete_assets)
    parser_delete.add_argument("-r", "--retention-file", required=True, type=Path,
                               help="File containing names of assets that will not be deleted")
    parser_delete.add_argument("-d", "--dry-run", action="store_true",
                               help="Dry run, don't actually make changes")
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Model utils Operations Class."""

import copy
import azureml.assets as assets
from typing import Tuple, Union
from azure.ai.ml import load_model, load_data, MLClient, operations as ops
from azure.ai.ml._utils._registry_utils import get_asset_body_for_registry_storage
from azureml.assets.util import logger
from azureml.assets.config import AssetType, Config, DataConfig, ModelConfig
from azureml.assets.util.util import resolve_from_file_for_asset
from azureml.assets.config import PathType
from azureml.assets.model.download_utils import CopyUpdater, copy_azure_artifacts, download_git_model
from azureml.assets.deployment_config import AssetVersionUpdate
from pathlib import Path


class RegistryUtils:
    """Registry utils."""

    RETRY_COUNT = 3

    def pluralize_asset_type(asset_type: Union[AssetType, str]) -> str:
        """Return pluralized asset type."""
        # Convert to string if enum
        if isinstance(asset_type, AssetType):
            asset_type = asset_type.value
        return f"{asset_type}s" if asset_type != "data" else asset_type

    def get_operations_from_type(asset_type: AssetType, ml_client: MLClient) -> Union[
                                 ops.ComponentOperations, ops.DataOperations, ops.EnvironmentOperations,
                                 ops.ModelOperations]:
        """Get MLCLient operations related to an asset type.

        Args:
            asset_type (AssetType): Asset type.
            ml_client (MLClient): ML client.

        Returns:
            Union[ops.ComponentOperations, ops.DataOperations, ops.EnvironmentOperations,
                ops.ModelOperations]: Operations object.
        """
        if asset_type == AssetType.COMPONENT:
            return ml_client.components
        elif asset_type == AssetType.DATA:
            return ml_client.data
        elif asset_type == AssetType.ENVIRONMENT:
            return ml_client.environments
        elif asset_type == AssetType.MODEL:
            return ml_client.models

    def publish_to_registry(ml_client: MLClient, extra_config: Config, asset_name: str, asset_version: str,
                            asset_type: assets.AssetType, temp_dir: Path, copy_updater: CopyUpdater = None,
                            output_level: str = "essential"):
        """Copy artifacts to registry storage.

        Args:
            ml_client (MLClient): ML client.
            extra_config (Config): ModelConfig or DataConfig object containing external storage info.
            asset_name (str): Asset name.
            asset_version (str): Asset version.
            asset_type (assets.AssetType): Asset type.
            temp_dir (Path): temp dir for asset operation.
            copy_updater (CopyUpdater): CopyUpdater object to update files during azcopy.
            output_level (str, optional): Output verbosity level parameter for azcopy. Defaults to "essential".
        """
        src_uri = extra_config.path.uri
        if extra_config.path.type == PathType.GIT:
            # download model locally (this is supported for models)
            src_uri = temp_dir
            logger.print(f"Cloning {asset_type} files from git {extra_config.path.uri}")
            success = download_git_model(extra_config.path.uri, src_uri)
            logger.print("Completed cloning")
            if not success:
                raise Exception(f"Cloning uri {extra_config.path.uri} failed")

        try:
            # copy to registry blobstorage (this is supported for models & data assets)
            logger.print("get data ref for registry storage upload")

            blob_uri, sas_uri = RegistryUtils.get_registry_data_reference(
                asset_name, asset_version, asset_type, ml_client
            )

            success = copy_azure_artifacts(src_uri=src_uri, dstn_uri=sas_uri, copy_updater=copy_updater,
                                           output_level=output_level)
            if not success:
                raise Exception("blobstorage copy failed.")
            logger.print(f"Successfully copied {asset_type.value} artifacts to registry storage")
            return blob_uri
        except Exception as e:
            raise Exception(f"Error in copying artifacts to registry storage. Error {e}")

    def get_registry_data_reference(asset_name: str, asset_version: str, asset_type: assets.AssetType,
                                    ml_client: MLClient) -> Tuple[str, str]:
        """Fetch data reference for asset in the registry."""
        asset_type_pluralized = RegistryUtils.pluralize_asset_type(asset_type)
        operations = RegistryUtils.get_operations_from_type(asset_type=asset_type, ml_client=ml_client)
        registry_name = operations._registry_name

        asset_id = f"azureml://registries/{registry_name}/{asset_type_pluralized}/" \
                   f"{asset_name}/versions/{asset_version}"
        logger.print(f"getting data reference for asset {asset_id}")
        for cnt in range(1, RegistryUtils.RETRY_COUNT + 1):
            try:
                response = RegistryUtils._get_temp_data_ref(asset_name, asset_version, asset_type, ml_client)
                blob_uri = response.blob_reference_for_consumption.blob_uri
                sas_uri = response.blob_reference_for_consumption.credential.additional_properties["sasUri"]
                if not blob_uri or not sas_uri:
                    raise Exception("Error fetching blob or SAS URI")
                return blob_uri, sas_uri
            except Exception as e:
                logger.log_error(f"Exception in fetching data reference. Try #{cnt}. Error: {e}")
        else:
            raise Exception(f"Unable to fetch data reference for asset {asset_id}")

    def _get_temp_data_ref(asset_name: str, asset_version: str, asset_type: assets.AssetType, ml_client: MLClient):
        asset_type_pluralized = RegistryUtils.pluralize_asset_type(asset_type)
        operations = RegistryUtils.get_operations_from_type(asset_type=asset_type, ml_client=ml_client)
        service_client = operations._service_client
        resource_group_name = operations._resource_group_name
        registry_name = operations._registry_name

        body = get_asset_body_for_registry_storage(registry_name, asset_type_pluralized, asset_name, asset_version)
        response = service_client.temporary_data_references.create_or_get_temporary_data_reference(
            name=asset_name,
            version=asset_version,
            resource_group_name=resource_group_name,
            registry_name=registry_name,
            body=body,
        )
        return response


class Asset:
    """Asset class."""

    def __init__(self, spec_path: Path, extra_config: Config, registry_name: str, temp_dir: Path,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
        """Initialize asset."""
        self._spec_path = spec_path
        self._extra_config = extra_config
        self._temp_dir = temp_dir
        self._copy_updater = copy_updater
        self._output_level = output_level


class DataAsset(Asset):
    """Asset class for data."""

    def __init__(self, spec_path: Path, data_config: DataConfig, registry_name: str, temp_dir: Path,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
        """Initialize data asset."""
        super().__init__(spec_path, data_config, registry_name, temp_dir, copy_updater, output_level)
        self._data_config = data_config

        try:
            self._data = load_data(spec_path)
        except Exception as e:
            logger.error(f"Error in loading data spec file at {spec_path}: {e}")
            return False

    def prepare_data(self, ml_client: MLClient):
        """Prepare data for publish."""
        data_registry_path = RegistryUtils.publish_to_registry(ml_client, self._data_config, self._data.name,
                                                               self._data.version, AssetType.DATA, self._temp_dir,
                                                               self._copy_updater, self._output_level)
        self._data.path = data_registry_path
        return self._data


class ModelAsset(Asset):
    """Asset class for model."""

    def __init__(self, spec_path: Path, model_config: ModelConfig, registry_name: str, temp_dir: Path,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
        """Initialize model asset."""
        super().__init__(spec_path, model_config, registry_name, temp_dir, copy_updater, output_level)
        self._model_config = model_config

        try:
            self._model = load_model(spec_path)
            self._model.description = model_config.description
            self._model.type = model_config.type.value
            if self._model.tags:
                self._model.tags = {k: resolve_from_file_for_asset(self._model_config, v)
                                    for k, v in self._model.tags.items()}
        except Exception as e:
            logger.error(f"Error in loading model spec file at {spec_path}: {e}")
            return False


class MLFlowModelAsset(ModelAsset):
    """Asset class for MLflow model."""

    MLMODEL_FILE_NAME = "MLmodel"
    MLFLOW_MODEL_PATH = "mlflow_model_folder"

    def __init__(self, spec_path: Path, model_config: ModelConfig, registry_name: str, temp_dir: Path,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
        """Initialize Mlflow model asset."""
        super().__init__(spec_path, model_config, registry_name, temp_dir, copy_updater, output_level)

    def prepare_model(self, ml_client: MLClient):
        """Prepare model for publish."""
        model_registry_path = RegistryUtils.publish_to_registry(ml_client, self._model_config, self._model.name,
                                                                self._model.version, AssetType.MODEL, self._temp_dir,
                                                                self._copy_updater, self._output_level)
        self._model.path = model_registry_path + "/" + MLFlowModelAsset.MLFLOW_MODEL_PATH
        return self._model


class TritonModelAsset(ModelAsset):
    """Asset class for Triton model."""

    def __init__(self, spec_path: Path, model_config: ModelConfig, registry_name: str, temp_dir: Path,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
        """Initialize Triton model asset."""
        super().__init__(spec_path, model_config, registry_name, temp_dir, copy_updater, output_level)

    def prepare_model(self, ml_client: MLClient):
        """Prepare model for publish."""
        model_registry_path = RegistryUtils.publish_to_registry(ml_client, self._model_config, self._model.name,
                                                                self._model.version, AssetType.MODEL, self._temp_dir,
                                                                self._copy_updater, self._output_level)
        self._model.path = model_registry_path
        return self._model


class CustomModelAsset(ModelAsset):
    """Asset class for custom model."""

    def __init__(self, spec_path: Path, model_config: ModelConfig, registry_name: str, temp_dir: Path,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
        """Initialize custom model asset."""
        super().__init__(spec_path, model_config, registry_name, temp_dir, copy_updater, output_level)

    def prepare_model(self, ml_client: MLClient):
        """Prepare model for publish."""
        model_registry_path = RegistryUtils.publish_to_registry(ml_client, self._model_config, self._model.name,
                                                                self._model.version, AssetType.MODEL, self._temp_dir,
                                                                self._copy_updater, self._output_level)
        self._model.path = model_registry_path
        return self._model


def prepare_model(spec_path: Path, model_config: ModelConfig, temp_dir: Path, ml_client: MLClient,
                  copy_updater: CopyUpdater = None, output_level: str = "essential"):
    """Prepare model for publish."""
    try:
        logger.print(f"Model type: {model_config.type}")
        registry_name = ml_client.models._registry_name
        if model_config.type == assets.ModelType.CUSTOM:
            model_asset = CustomModelAsset(spec_path, model_config, registry_name, temp_dir,
                                           copy_updater, output_level)
        elif model_config.type == assets.ModelType.MLFLOW:
            model_asset = MLFlowModelAsset(spec_path, model_config, registry_name, temp_dir,
                                           copy_updater, output_level)
        elif model_config.type == assets.ModelType.TRITON:
            model_asset = TritonModelAsset(spec_path, model_config, registry_name, temp_dir,
                                           copy_updater, output_level)
        else:
            logger.log_error(f"Model type {model_config.type.value} not supported")
            return False
        model = model_asset.prepare_model(ml_client)
        return model, True
    except Exception as e:
        logger.log_error(f"prepare model failed for {spec_path}. Error {e}")
        return None, False


def prepare_data(spec_path: Path, data_config: DataConfig, temp_dir: Path, ml_client: MLClient,
                 copy_updater: CopyUpdater = None, output_level: str = "essential"):
    """Prepare data for publish."""
    try:
        registry_name = ml_client.data._registry_name
        data_asset = DataAsset(spec_path, data_config, registry_name, temp_dir, copy_updater, output_level)
        data = data_asset.prepare_data(ml_client)
        return data, True
    except Exception as e:
        logger.log_error(f"prepare data failed for {spec_path}. Error {e}")
        return None, False


def update_metadata(
    name: str,
    version: str,
    update: AssetVersionUpdate,
    ml_client: MLClient,
    asset_type: assets.AssetType,
    allow_no_op_update: bool = False,
):
    """Update the mutable metadata of an already registered asset."""
    try:
        # Get existing asset
        operations = RegistryUtils.get_operations_from_type(asset_type=asset_type, ml_client=ml_client)
        asset = operations.get(name=name, version=version)
        need_update = False

        # Update tags
        if update.tags:
            updated_tags = copy.deepcopy(asset.tags)

            if update.tags.replace is not None:
                updated_tags = update.tags.replace
            else:
                if update.tags.add is not None:
                    for k, v in update.tags.add.items():
                        updated_tags[k] = v
                if update.tags.delete is not None:
                    for k in update.tags.delete:
                        updated_tags.pop(k, None)

            if updated_tags != asset.tags:
                logger.print("tags has been updated.")
                asset.tags = updated_tags
                need_update = True

        # Update properties
        if update.properties:
            updated_properties = copy.deepcopy(asset.properties)
            if update.properties.add is not None:
                for k, v in update.properties.add.items():
                    if k in asset.properties and asset.properties[k] != v:
                        raise Exception(f"Value of property {k} for {asset_type.value} {name} cannot "
                                        f"be replaced to {v} without increasing the version.")
                    updated_properties[k] = v

            if updated_properties != asset.properties:
                logger.print("properties has been updated")
                asset.properties = updated_properties
                need_update = True

        # Update description
        if update.description is not None and asset.description != update.description:
            logger.print("description has been updated")
            asset.description = update.description
            need_update = True

        if allow_no_op_update or need_update:
            operations.create_or_update(asset)
            logger.print(f"{asset_type.value.capitalize()} metadata updated successfully for {name}")
        else:
            logger.print(f"No update found for {asset_type.value} {name}. Skipping")
    except Exception as e:
        logger.log_error(f"Failed to update metadata for {asset_type.value} : {name} : {e}")

    # Archive or restore asset based on stage
    try:
        if update.stage == "Archived":
            logger.print(f"Archiving {asset_type.value} {name} version {version}")
            operations.archive(name=name, version=version)
        elif update.stage == "Active":
            logger.print(f"Restoring {asset_type.value} {name} version {version}")
            operations.restore(name=name, version=version)
    except Exception as e:
        logger.log_error(f"Failed to archive or restore {asset_type.value} {name} version {version}: {e}")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Functions for replacing template tags."""


TAG_PREFIX = "{{"
TAG_SUFFIX = "}}"
TAG_SEPARATOR = "."


def _process_tag(tag: str, data: dict) -> str:
    """Replace tag with data value, or leave intact if not found in data.

    Args:
        tag (str): Template tag, including prefix and suffix.
        data (dict): Data to replace tag with, may include nested dictionaries.

    Returns:
        str: The data value or original template tag if not found.
    """
    inside_tag = tag[len(TAG_PREFIX):-len(TAG_SUFFIX)].strip()
    tag_segments = inside_tag.split(TAG_SEPARATOR)
    data_context = data
    for tag_segment in tag_segments:
        val = data_context.get(tag_segment)
        if isinstance(val, dict):
            data_context = val
        elif val is not None:
            return val
        else:
            return tag


def render(template: str, data: dict) -> str:
    """Replace template tags with data, or leave them intact if not found in data.

    Args:
        template (str): Template that may include tags to replace.
        data (dict): Data to replace tags with, may include nested dictionaries.

    Returns:
        str: Template with resolved tags replaced with their values.
    """
    # Split template into a list of strings separated by template tags
    template_components = []
    start = 0
    while start < len(template) - 1:
        # Find a template tag
        tag_start = template.find(TAG_PREFIX, start)
        tag_suffix_start = template.find(TAG_SUFFIX, tag_start) if tag_start != -1 else -1
        if tag_start == -1 or tag_suffix_start == -1:
            # No template tag found, store remainder
            template_components.append(template[start:])
            break

        # Compute end position of tag
        tag_end = tag_suffix_start + len(TAG_SUFFIX) - 1

        if tag_start > start:
            # Store string before the template tag
            template_components.append(template[start:tag_start])
        # Store template tag
        template_components.append(template[tag_start:tag_end + 1])
        start = tag_end + 1

    # Process template tags
    for i, template_component in enumerate(template_components):
        # Skip non-tags
        if not template_component.startswith(TAG_PREFIX):
            continue

        # Replace tag with data value or leave intact
        template_components[i] = _process_tag(template_component, data)

    return "".join(template_components)
